local EffectTemplate = import("/lua/effecttemplates.lua")
local explosion = import("/lua/defaultexplosions.lua")
local Entity = import("/lua/sim/entity.lua").Entity
local MobileUnit = import("/lua/sim/units/mobileunit.lua").MobileUnit
local MobileUnitOnCreate = MobileUnit.OnCreate
local MobileUnitOnMotionVertEventChange = MobileUnit.OnMotionVertEventChange
local utilities = import("/lua/utilities.lua")
---@class SubUnit : MobileUnit
SubUnit = ClassUnit(MobileUnit) {
    -- Use default spark effect until underwater damaged states are made
    FxDamage1 = { EffectTemplate.DamageSparks01 },
    FxDamage2 = { EffectTemplate.DamageSparks01 },
    FxDamage3 = { EffectTemplate.DamageSparks01 },

    -- DESTRUCTION PARAMS
    ShowUnitDestructionDebris = false,
    DeathThreadDestructionWaitTime = 0,

    ---@param self SubUnit
    ---@param spec any
    OnCreate = function(self, spec)
        MobileUnitOnCreate(self, spec)

        -- submarines do not make a sound by default, we want them to make sound so we use an entity as source instead
        self.SoundEntity = Entity()
        self.Trash:Add(self.SoundEntity)
        Warp(self.SoundEntity, self:GetPosition())
        self.SoundEntity:AttachTo(self, -1)
    end,
    OnKilled = function(self, instigator, type, overkillRatio)
        local layer = self:GetCurrentLayer()
        self:DestroyIdleEffects()
        local bp = self:GetBlueprint()
        local numBones = self:GetBoneCount() - 1
		
        if (layer == 'Water' or layer == 'Seabed' or layer == 'Sub') and bp.Display.AnimationDeath then
			self.SinkExplosionThread = self:ForkThread(self.ExplosionThread)
            self.SinkThread = self:ForkThread(self.SinkingThread)
        end
        if (layer == 'Sub') then
			for i = 0, numBones do
				boner = Random(0, numBones)
				explosion.CreateDefaultHitExplosionAtBone( self, boner, -1.0 )
			end
			self.SubThread = self:ForkThread(self.SubThread)
		end
		MobileUnit.OnKilled(self, instigator, type, overkillRatio)
    end,

    SubThread = function(self)
		local numBones = self:GetBoneCount() - 1    
		while true do 
			boner = Random(0, numBones)
			explosion.CreateDefaultHitExplosionAtBone( self, boner, -1.0 )
			WaitSeconds(0.4)	
        end
    end,
	
	ExplosionThread = function(self)
        local maxcount = Random(17,20) --max number of above surface explosions. timed to animation
        local d = 0 --delay offset after surface explosions cease
        local sx, sy, sz = self:GetUnitSizes()
        local vol = sx * sy * sz

        local volmin = 1.5
        local volmax = 15
        local scalemin = 1
        local scalemax = 3
        local t = (vol-volmin)/(volmax-volmin)
        local rs = scalemin + (t * (scalemax-scalemin))
        if rs < scalemin then
            rs = scalemin
        elseif rs > scalemax then
            rs = scalemax
        end
        local army = self:GetArmy()

        CreateEmitterAtEntity(self,army,'/effects/emitters/destruction_underwater_explosion_flash_01_emit.bp'):ScaleEmitter(rs)
        CreateEmitterAtEntity(self,army,'/effects/emitters/destruction_underwater_explosion_splash_02_emit.bp'):ScaleEmitter(rs)
        CreateEmitterAtEntity(self,army,'/effects/emitters/destruction_underwater_explosion_surface_ripples_01_emit.bp'):ScaleEmitter(rs)

        while true do
            local rx, ry, rz = self:GetRandomOffset(1)
            local rs = Random(vol/2, vol*2) / (vol*2)
            CreateEmitterAtEntity(self,army,'/effects/emitters/destruction_underwater_explosion_flash_01_emit.bp'):ScaleEmitter(rs):OffsetEmitter(rx, ry, rz)
            CreateEmitterAtEntity(self,army,'/effects/emitters/destruction_underwater_explosion_splash_01_emit.bp'):ScaleEmitter(rs):OffsetEmitter(rx, ry, rz)

            d = d + 1 --increase delay offset
            local rd = Random(30,70) / 10
            WaitTicks(rd + d)
        end
    end,
    ---@param self Unit
    ---@param new VerticalMovementState
    ---@param old VerticalMovementState
    OnMotionVertEventChange = function(self, new, old)
        MobileUnitOnMotionVertEventChange(self, new, old)

        -- only temporarily remove/add dive command if the submarine should have a dive command
        if self.Blueprint.General.CommandCaps.RULEUCC_Dive then
            if new == 'Up' or new == 'Down' then
                self:RemoveCommandCap("RULEUCC_Dive")
                self:RequestRefreshUI()
            end

            if new == 'Top' or new == 'Bottom' then
                self:AddCommandCap("RULEUCC_Dive")
                self:RequestRefreshUI()
            end
        end
    end,
}
