
local TIFTacticalNuke = import("/lua/terranprojectiles.lua").TIFTacticalNuke
local EffectTemplate = import("/lua/effecttemplates.lua")

local VisionMarkerOpti = import("/lua/sim/vizmarker.lua").VisionMarkerOpti

-- upvalue scope for performance
local WaitTicks = WaitTicks
local ForkThread = ForkThread
local DamageArea = DamageArea

--- used by uel0001
---@class TIFMissileNukeCDR : TIFTacticalNuke
---@field Armed boolean
TIFMissileNukeCDR = ClassProjectile(TIFTacticalNuke) {
    -- BeamName = '/effects/emitters/missile_exhaust_fire_beam_01_emit.bp',
    BeamName = '/effects/emitters/missile_exhaust_fire_beam_06_emit.bp',
    InitialEffects = { '/effects/emitters/nuke_munition_launch_trail_02_emit.bp',
    '/mods/Firey Explosion Mod FAF/hook/effects/Emitters/tnuketakes.bp', },
    ThrustEffects = { '/effects/emitters/nuke_munition_launch_trail_04_emit.bp', },
    LaunchEffects = {
        '/effects/emitters/nuke_munition_launch_trail_03_emit.bp',
        '/effects/emitters/nuke_munition_launch_trail_05_emit.bp',
        '/mods/Firey Explosion Mod FAF/hook/effects/Emitters/tnukelenses.bp',
		'/mods/Firey Explosion Mod FAF/hook/effects/Emitters/tnuketrails.bp',
    },

    DebrisBlueprints = {
        '/effects/Entities/TacticalDebris02/TacticalDebris02_proj.bp',
    },

    -- reduce due to distance
    LaunchTicks = 6,
    HeightDistanceFactor = 10,
    FinalBoostAngle = 60,

    ---@param self TIFMissileNukeCDR
    ---@param inWater boolean
    OnCreate = function(self, inWater)
        TIFTacticalNuke.OnCreate(self)
        if not inWater then
            self:SetDestroyOnWater(true)
        end

        self.MoveThread = self.Trash:Add(ForkThread(self.MovementThread, self))
        self.effectEntityPath = '/effects/Entities/UEFNukeEffectController02/UEFNukeEffectController02_proj.bp'
        self:LauncherCallbacks()
        self.Armed = false

        local army = self.Army
        local trash = self.Trash
        for _, effect in self.InitialEffects do
            trash:Add(CreateAttachedEmitter(self, -1, army, effect))
        end

        local launchEffects = {}
        for _, effect in self.ThrustEffects do
            table.insert(launchEffects, trash:Add(CreateAttachedEmitter(self, -1, army, effect)))
        end

        for _, effect in self.LaunchEffects do
            table.insert(launchEffects, trash:Add(CreateAttachedEmitter(self, -1, army, effect)))
        end

        trash:Add(ForkThread(self.ArmMissileThread, self, launchEffects))
    end,
    CreateEffects = function( self, EffectTable, army, scale)
        for k, v in EffectTable do
            self.Trash:Add(CreateAttachedEmitter(self, -1, army, v):ScaleEmitter(scale))
        end
    end,

    ---@param self TIFMissileNukeCDR
    OnExitWater = function(self)
        TIFTacticalNuke.OnExitWater(self)
        self:SetDestroyOnWater(true)
    end,

    ---@param self TIFMissileNukeCDR
    ---@param launchEffects moho.IEffect[]
    ArmMissileThread = function(self, launchEffects)
        WaitTicks(7)
        for k, effect in launchEffects do
            effect:ScaleEmitter(0.9)
        end

        WaitTicks(7)
        for k, effect in launchEffects do
            effect:ScaleEmitter(0.7)
        end

        WaitTicks(7)
        for k, effect in launchEffects do
            effect:ScaleEmitter(0.5)
        end

        WaitTicks(7)
        for k, effect in launchEffects do
            effect:ScaleEmitter(0.3)
        end

        -- indicates to the player(s) that the billy is now armed
        self:CreateDebris()

        self.Armed = true
    end,

    ---@param self TIFMissileNukeCDR
    ---@param targetType string
    ---@param targetEntity Prop|Unit
    OnImpact = function(self, targetType, targetEntity)
        local position = self:GetPosition()

        if self.Armed then
            -- create vision
            local marker = VisionMarkerOpti({ Owner = self })
            marker:UpdatePosition(position[1], position[3])
            marker:UpdateDuration(9)
            marker:UpdateIntel(self.Army, 12, 'Vision', true)

            TIFTacticalNuke.OnImpact(self, targetType, targetEntity)
        else
            -- default tactical explosion
            self:CreateDebris()
            self:Destroy()
        end
    end,

    --- Called by Lua to process the overriden damage logic of TIFMissileNukeCDR.
    -- Similar to nuke damage logic, except it does not bypass shields.
    -- @param self TIFMissileNukeCDR
    -- @param instigator The launcher, and if it doesn't exist, the projectile itself
    -- @param DamageData The damage data passed by the weapon
    -- @param targetEntity The entity we hit, is nil if we hit terrain
    -- @param cachedPosition A cached position that is passed to prevent table allocations, can not be used in fork threads and / or after a yield statement
    ---@param self TIFMissileNukeCDR
    ---@param instigator UEL0001 | TIFMissileNukeCDR
    ---@param DamageData table
    ---@param targetEntity Unit | Prop
    ---@param cachedPosition Vector
    DoDamage = function(self, instigator, DamageData, targetEntity, cachedPosition)
        local InnerRing = self.InnerRing
        local OuterRing = self.OuterRing
        DamageArea(instigator, cachedPosition, InnerRing.Radius, InnerRing.Damage, 'Normal', true, true)
        DamageArea(instigator, cachedPosition, OuterRing.Radius, OuterRing.Damage, 'Normal', true, true)
    end,
    MovementThread = function(self)   
        local army = self:GetArmy()
        local target = self:GetTrackingTarget()
        local launcher = self:GetLauncher()            
        self.CreateEffects( self, self.InitialEffects, army, 1 )      
        self.WaitTime = 0.1
        self:SetTurnRate(8)
        WaitSeconds(0.3)   
        self.CreateEffects( self, self.LaunchEffects, army, 1 )
        self.CreateEffects( self, self.ThrustEffects, army, 1 )        
        while not self:BeenDestroyed() do
            self:SetTurnRateByDist()
            WaitSeconds(self.WaitTime)
        end
    end,

    SetTurnRateByDist = function(self)
        local dist = self:GetDistanceToTarget()
        if dist > 50 then        
            --Freeze the turn rate as to prevent steep angles at long distance targets
            WaitSeconds(2)
            self:SetTurnRate(20)
        elseif dist > 128 and dist <= 213 then
			-- Increase check intervals
			self:SetTurnRate(30)
			WaitSeconds(1.5)
            self:SetTurnRate(30)
        elseif dist > 43 and dist <= 107 then
		    -- Further increase check intervals
            WaitSeconds(0.3)
            self:SetTurnRate(75)
		elseif dist > 0 and dist <= 43 then
			-- Further increase check intervals            
            self:SetTurnRate(200)   
            KillThread(self.MoveThread)         
        end
    end,        

    GetDistanceToTarget = function(self)
        local tpos = self:GetCurrentTargetPosition()
        local mpos = self:GetPosition()
        local dist = VDist2(mpos[1], mpos[3], tpos[1], tpos[3])
        return dist
    end,
    
    OnEnterWater = function(self)
        TIFMissileNuke.OnEnterWater(self)
        self:SetDestroyOnWater(true)
    end,
}
TypeClass = TIFMissileNukeCDR
